#define _POSIX_C_SOURCE 200809L

#include <common/common.h>
#include <lektor/database.h>

#include <strings.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>

static inline int
__check_sticker_type(const char *type)
{
    return ! ( STR_MATCH(type, "kara") || STR_MATCH(type, "plt") );
}

bool
database_search_database_init(volatile sqlite3 *db, struct lkt_search *ret)
{
    RETURN_UNLESS(ret, "Exit because return pointer is NULL", false);
    static const char *SQL_STMT_TEMPLATE =
        "WITH content AS ("
        " SELECT kara.id AS id, string, LENGTH(CAST(kara.id AS TEXT)) AS len"
        " FROM kara WHERE %s LIKE ?)"
        "SELECT id, string, (SELECT MAX(len) FROM content)"
        "FROM content LIMIT %d OFFSET %d;";
    char SQL_STMT[LKT_MAX_SQLITE_STATEMENT];
    ret->type = lkt_search_database;

    /* Search part */
    safe_snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT, SQL_STMT_TEMPLATE,
                  ret->ka_uri.column_name, ret->msg_count, ret->continuation);
    SQLITE_PREPARE(db, ret->stmt, SQL_STMT, error);
    if (!ret->ka_uri.is_int) {
        SQLITE_BIND_TEXT(db, ret->stmt, 1, ret->ka_uri.value, error);
        LOG_DEBUG("DB", "%s", "Uri was a text");
    } else {
        SQLITE_BIND_INT(db, ret->stmt, 1, (int) ret->ka_uri.id, error);
        LOG_DEBUG("DB", "%s", "Uri was a size_t");
    }
    ret->db = db;
    return true;
error:
    sqlite3_finalize(ret->stmt);
    return false;
}

bool
database_search_playlist_init(volatile sqlite3 *db, struct lkt_search *ret)
{
    /* Sql queries definitions */
    char SQL_STMT[LKT_MAX_SQLITE_STATEMENT];
    static const char *SQL_TEMPLATE =
        "WITH content AS ("
        " SELECT kara.id AS id, string, LENGTH(CAST(kara.id AS TEXT)) AS len"
        " FROM kara_playlist"
        " JOIN playlist ON playlist.id = playlist_id"
        "               AND playlist.name COLLATE nocase = ?"
        " JOIN kara ON kara.id = kara_id"
        "           AND %s LIKE ?) "
        "SELECT id, string, (SELECT MAX(len) FROM content) "
        "FROM content LIMIT %d OFFSET %d;";
    static const char *SQL =
        "WITH content AS ("
        " SELECT kara.id AS id, string, LENGTH(CAST(kara.id AS TEXT)) AS len"
        " FROM kara_playlist"
        " JOIN playlist ON playlist.id = playlist_id"
        "               AND playlist.name COLLATE NOCASE = ?"
        " JOIN kara ON kara.id = kara_id) "
        "SELECT id, string, (SELECT MAX(len) FROM content) "
        "FROM content LIMIT %d OFFSET %d;";

    /* Common part */
    RETURN_UNLESS(ret, "Exit because return pointer is NULL", false);
    ret->type   = lkt_search_playlist;
    ret->db     = db;
    if (ret->ka_uri.type == uri_null)
        goto uri_is_null;

    /* Filter with an uri */
    LOG_INFO("DB", "List plt '%s' with non null uri", ret->plt_name);
    safe_snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT, SQL_TEMPLATE,
                  ret->ka_uri.column_name, ret->msg_count, ret->continuation);
    SQLITE_PREPARE(db, ret->stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, ret->stmt, 1, ret->plt_name, error);
    if (!ret->ka_uri.is_int) {
        SQLITE_BIND_TEXT(db, ret->stmt, 2, ret->ka_uri.value, error);
        LOG_DEBUG("DB", "%s", "Uri was a text");
    } else {
        SQLITE_BIND_INT(db, ret->stmt, 2, (int) ret->ka_uri.id, error);
        LOG_DEBUG("DB", "%s", "Uri was a size_t");
    }
    return true;

    /* Don't filter */
uri_is_null:
    LOG_INFO("DB", "List plt '%s' with null uri", ret->plt_name);
    safe_snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT, SQL, ret->msg_count, ret->continuation);
    SQLITE_PREPARE(db, ret->stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, ret->stmt, 1, ret->plt_name, error);
    return true;

    /* Error */
error:
    sqlite3_finalize(ret->stmt);
    return false;
}

bool
database_search_listplaylist_init(volatile sqlite3 *db, struct lkt_search *ret)
{
    static const char *SQL = "SELECT name FROM playlist LIMIT %d OFFSET %d;";
    char SQL_STMT[LKT_MAX_SQLITE_STATEMENT];
    safe_snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT, SQL, ret->msg_count, ret->continuation);
    ret->type   = lkt_search_listplaylist;
    ret->db     = db;
    SQLITE_PREPARE(db, ret->stmt, SQL_STMT, error);
    LOG_INFO("DB", "List playlists with LIMIT %d and OFFSET %d",
             ret->msg_count, (int) ret->continuation);
    return true;
error:
    sqlite3_finalize(ret->stmt);
    return false;
}

bool
database_search_sticker_init(volatile sqlite3 *db, struct lkt_search *ret)
{
    if (__check_sticker_type(ret->st_type)) {
        LOG_ERROR("DB", "Type '%s' is invalid", ret->st_type);
        return false;
    }

    static const char *SQL_all_types =
        "SELECT name, sts.id, value FROM 'stickers' JOIN "
        "( SELECT id, sticker, value, 'kara' FROM 'stickers.kara'"
        "  UNION"
        "  SELECT id, sticker, value , 'plt' FROM 'stickers.plt'"
        ") AS sts"
        "ON sts.sticker = 'stickers'.id";
    static const char *SQL_one_type =
        "SELECT name, sts.id, value, '%s' "
        "FROM 'stickers.%s' AS sts "
        "JOIN 'stickers' "
        "ON 'stickers'.id = sts.sticker";
    char SQL[LKT_MAX_SQLITE_STATEMENT];
    ret->type = lkt_search_sticker;

    if (ret->st_type == NULL)
        memcpy(SQL, SQL_all_types, strlen(SQL_all_types) + 1);
    else
        sprintf(SQL, SQL_one_type, ret->st_type, ret->st_type);
    if (ret->st_uri != 0)
        sprintf(SQL + strlen(SQL), " AND sts.id = %d", ret->st_uri);
    if (ret->st_op != 0)
        sprintf(SQL + strlen(SQL), " AND sts.value %s %d",
                ret->st_op == '>' ? ">=" : ret->st_op == '<' ? "<=" : "=",
                ret->st_value);
    strcat(SQL, !ret->st_name ? ";" : " AND name = ?;");

    SQLITE_PREPARE(db, ret->stmt, SQL, error);
    if (ret->st_name)
        SQLITE_BIND_TEXT(db, ret->stmt, 1, ret->st_name, error);
    return true;
error:
    sqlite3_finalize(ret->stmt);
    return false;
}

bool
database_search_queue_init(volatile sqlite3 *db, struct lkt_search *ret)
{
    RETURN_UNLESS(ret, "Exit because return pointer is NULL", false);
    static const char *SQL_STMT_TEMPLATE =
        "WITH content AS ("
        " SELECT"
        "  kara_id AS id,"
        "  string,"
        "  LENGTH(CAST(kara_id AS TEXT)) AS len"
        " FROM queue"
        " JOIN kara "
        " ON kara_id = kara.id AND %s LIKE ?)"
        "SELECT id, string, (SELECT MAX(len) FROM content)"
        "FROM content LIMIT %d OFFSET %d;";
    char SQL_STMT[LKT_MAX_SQLITE_STATEMENT];
    ret->type = lkt_search_queue;

    /* Search part */
    safe_snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT, SQL_STMT_TEMPLATE,
                  ret->ka_uri.column_name, ret->msg_count, ret->continuation);
    SQLITE_PREPARE(db, ret->stmt, SQL_STMT, error);
    if (!ret->ka_uri.is_int) {
        SQLITE_BIND_TEXT(db, ret->stmt, 1, ret->ka_uri.value, error);
        LOG_DEBUG("DB", "%s", "Uri was a text");
    } else {
        SQLITE_BIND_INT(db, ret->stmt, 1, (int) ret->ka_uri.id, error);
        LOG_DEBUG("DB", "%s", "Uri was a size_t");
    }
    ret->db = db;
    return true;
error:
    sqlite3_finalize(ret->stmt);
    return false;
}

bool
database_search_iter(struct lkt_search *item)
{
    const char *sql_row, *type;
    int id, code, id_len;
    RETURN_UNLESS(item, "Exit because item is NULL, end iterations", false);
    GOTO_UNLESS(item->call, "NULL call function, terminate the search", end);
    code = sqlite3_step(item->stmt);

    if (code == SQLITE_DONE)
        goto end;

    if (code != SQLITE_ROW) {
        LOG_ERROR("DB", "Step failed, expected a ROW or a DONE: %s",
                  sqlite3_errmsg((sqlite3 *) item->db));
        goto end;
    }

    switch (item->type) {
    case lkt_search_database:
    case lkt_search_queue:
    case lkt_search_playlist:
        id      = sqlite3_column_int  (item->stmt, 0);
        sql_row = sqlite3_column_chars(item->stmt, 1);
        id_len  = sqlite3_column_int  (item->stmt, 2);
        return ((lkt_search_database_func) item->call)
               (item->srv, item->c, id, id_len, sql_row);
    case lkt_search_sticker:
        sql_row = sqlite3_column_chars(item->stmt, 0);  /* Name */
        id      = sqlite3_column_int  (item->stmt, 1);  /* Id   */
        code    = sqlite3_column_int  (item->stmt, 2);  /* Val  */
        type    = sqlite3_column_chars(item->stmt, 3);  /* Type */
        return ((lkt_search_sticker_func) item->call)
               (item->srv, item->c, sql_row, type, id, code);
    case lkt_search_listplaylist:
        sql_row = sqlite3_column_chars(item->stmt, 0);  /* Name */
        return ((lkt_search_listplaylist_func) item->call)
                (item->srv, item->c, sql_row);
    default:
        LOG_WARN("DB", "Search type %d is not implemented", item->type);
        goto end;
    }

end:
    sqlite3_finalize(item->stmt);
    return false;
}

bool
database_kara_by_id(volatile sqlite3 *db, int id, struct kara_metadata *res,
                    char filepath[PATH_MAX])
{
    static const char *SQL =
        "SELECT song_name, source_name, category, language, author_name, "
        "song_type, song_number, file_path FROM kara WHERE id = ?;";
    sqlite3_stmt *stmt = 0;
    int ret = false;

    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_BIND_INT(db, stmt, 1, id, error);
    SQLITE_STEP_ROW(db, stmt, error);
    if (!res)
        goto no_metadata;
    strncpy(res->song_name,   sqlite3_column_chars(stmt, 0), LEKTOR_TAG_MAX - 1);
    strncpy(res->source_name, sqlite3_column_chars(stmt, 1), LEKTOR_TAG_MAX - 1);
    strncpy(res->category,    sqlite3_column_chars(stmt, 2), LEKTOR_TAG_MAX - 1);
    strncpy(res->language,    sqlite3_column_chars(stmt, 3), LEKTOR_TAG_MAX - 1);
    strncpy(res->author_name, sqlite3_column_chars(stmt, 4), LEKTOR_TAG_MAX - 1);
    strncpy(res->song_type,   sqlite3_column_chars(stmt, 5), LEKTOR_TAG_MAX - 1);
    res->song_number = sqlite3_column_int(stmt, 6);
no_metadata:
    strncpy(filepath, sqlite3_column_chars(stmt, 7), PATH_MAX - 1);
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}
