#define _POSIX_C_SOURCE 200809L

#include <common/common.h>
#include <common/stack.h>
#include <lektor/database.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <pthread.h>

extern unsigned char ___src_database_disk_sql[];
extern unsigned char ___src_database_memory_sql[];

static const char *__pragmas =
    "PRAGMA case_sensitive_like = false;\n"
    "PRAGMA encoding            = 'UTF-8';\n"
    "PRAGMA synchronous         = 0;\n"
    "PRAGMA journal_mode        = off;\n";

#define INVALID_CHARS_DBPATH    ":?!'\""
#define HEAP_LIMIT_SOFT         100 * 1024 * 1024
#define HEAP_LIMIT_HARD         150 * 1024 * 1024

struct named_db {
    volatile sqlite3 *db;
    const char *name;
};
static struct stack db_stack = {0};
static int atexited = 0;
static pthread_mutex_t mtx = PTHREAD_MUTEX_INITIALIZER;

static inline void
__dec(volatile sqlite3 *db, const char *name)
{
    char SQL[LKT_MAX_SQLITE_STATEMENT];
    safe_snprintf(SQL, LKT_MAX_SQLITE_STATEMENT,
                  "UPDATE %s.misc SET opened = 0;", name);
    SQLITE_EXEC(db, SQL, error);
    LOG_INFO("DB", "Dec open count on db '%s'", name);
    return;
error:
    LOG_ERROR("DB", "%s", "Faield to close database");
}

void
database_close_all(void)
{
    LOG_INFO("GARBAGE", "%s", "Closing all db");
    if (pthread_mutex_trylock(&mtx))
        return;
    struct named_db *item;
    while (!stack_empty(&db_stack)) {
        stack_pop(&db_stack, (void **) &item);
        __dec(item->db, item->name);
        free((void *) item->name);
        free(item);
    }
    LOG_INFO("GARBAGE", "%s", "All db closed");
    stack_free(&db_stack);
}

static inline void
__inc(volatile sqlite3 *db, const char *name, bool check)
{
    SQLITE_EXEC(db, "UPDATE misc SET opened = (SELECT opened + 1 FROM misc);",
                error);

    if (!atexited) {
        atexited = 1;
        if (stack_new(&db_stack))
            exit(EXIT_FAILURE);
    }

    struct named_db *item = malloc(sizeof(struct named_db));
    if (!item)
        exit(EXIT_FAILURE);
    item->name = strdup(name);
    item->db   = db;
    if (stack_push(&db_stack, item))
        exit(EXIT_FAILURE);
    return;
error:
    LOG_ERROR("DB", "%s", "Database already in use");
    if (check)
        exit(EXIT_FAILURE);
    __dec(db, name);
}

int
is_sql_str_invalid(const char *str)
{
    size_t len = strlen(str);
    return strcspn(str, INVALID_CHARS_DBPATH) != len;
}

bool
database_new(volatile sqlite3 **db)
{
    *db = NULL;
    static int flags = SQLITE_OPEN_READWRITE    |
                       SQLITE_OPEN_SHAREDCACHE  |
                       SQLITE_OPEN_NOFOLLOW     |
                       SQLITE_OPEN_FULLMUTEX;
    if (SQLITE_OK != sqlite3_enable_shared_cache(1)     ||
        sqlite3_soft_heap_limit64(HEAP_LIMIT_SOFT) < 0  ||
        sqlite3_hard_heap_limit64(HEAP_LIMIT_HARD) < 0  ||
        SQLITE_OK != sqlite3_open_v2(":memory:", (sqlite3 **) db, flags, NULL))
        return false;
    SQLITE_EXEC(*db, (const char *) ___src_database_memory_sql, err_not_init);
    SQLITE_EXEC(*db, __pragmas, err_pragma);
    return true;
err_not_init:
    *db = NULL;
    return false;
err_pragma:
    sqlite3_close((sqlite3 *) *db);
    *db = NULL;
    return false;
}

static inline bool
__attach(volatile sqlite3 *db, const char *name, const char *path)
{
    static const char SQL_ATTACH_TEMPLATE[] = "ATTACH '%s' AS '%s';";
    char SQL_ATTACH[LKT_MAX_SQLITE_STATEMENT];
    bool ret = false;

    safe_snprintf(SQL_ATTACH, LKT_MAX_SQLITE_STATEMENT, SQL_ATTACH_TEMPLATE,
                  path, name);
    SQLITE_EXEC(db, SQL_ATTACH, err_no_attach);
    LOG_INFO("DB", "Attached database '%s' with path '%s'", name, path);
    ret = true;
err_no_attach:
    return ret;
}

static inline bool
__detach(volatile sqlite3 *db, const char *name)
{
    static const char SQL_DETACH_TEMPLATE[] = "DETACH '%s';\n";
    char SQL_DETACH[LKT_MAX_SQLITE_STATEMENT];
    bool ret = false;

    safe_snprintf(SQL_DETACH, LKT_MAX_SQLITE_STATEMENT,
                  SQL_DETACH_TEMPLATE, name);
    SQLITE_EXEC(db, SQL_DETACH, err_no_detach);
    LOG_INFO("DB", "Detached database '%s'", name);
    ret = true;
err_no_detach:
    return ret;
}

static inline bool
__check_schema(volatile sqlite3 *db)
{
    static const char *SQL =
        "SELECT COUNT(name)"
        " FROM sqlite_master"
        " WHERE type='table' AND name IN"
        " ('kara', 'playlist', 'kara_playlist', 'queue', 'users', 'stickers',"
        "  '''stickers.kara''', '''stickers.plt''', 'updates', 'misc', "
        "  'kara_type', 'kara_category', 'language');";
    sqlite3_stmt *stmt = NULL;
    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_STEP_ROW(db, stmt, error);
    if (9 != sqlite3_column_int(stmt, 0))
        return true;
error:
    LOG_ERROR("DB", "%s", "Schema is invalid: missing tables");
    return false;
}

static inline bool
__is_attached(volatile sqlite3 *db, const char *name)
{
    static const char *SQL_STMT =
        "SELECT name FROM pragma_database_list WHERE name = ?;\n";
    sqlite3_stmt *stmt = 0;
    bool ret = false;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, name, error);
    SQLITE_STEP_ROW(db, stmt, error);
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_open(volatile sqlite3 *db, const char *dbpath, bool check)
{
    bool retry = false;
    if (is_sql_str_invalid(dbpath)) {
        LOG_ERROR("DB", "The database path '%s' is invalid", dbpath);
        return false;
    }

    /* Try the simple attach, for already initialized database */
retry:
    if (!__attach(db, PROTECTED_DATABASE, dbpath))
        goto init;
    if (!__check_schema(db))
        goto init;
    __inc(db, PROTECTED_DATABASE, check);
    return true;

    /* Need init */
init:
    if (retry) {
        LOG_ERROR("DB", "Retry to init database '%s', this is an error",
                  dbpath);
        return false;
    }
    retry = true;
    LOG_WARN("DB", "Database '%s' is not correctly initialized, init it",
             dbpath);
    __detach(db, PROTECTED_DATABASE) ;
    if (database_init(dbpath))
        goto retry;
    else {
        LOG_ERROR("DB", "Failed to init database '%s'", dbpath);
        return false;
    }
}

bool
database_attach(volatile sqlite3 *db, const char *name, const char *dbpath)
{
    RETURN_UNLESS(strcasecmp(PROTECTED_DATABASE, name), " The database "
                  PROTECTED_DATABASE " is protected, can't attach a "
                  "database with the same name", false);

    if (is_sql_str_invalid(name)) {
        LOG_ERROR("DB", "The database name '%s' is invalid", name);
        return false;
    }

    if (__is_attached(db, name)) {
        LOG_ERROR("DB", "The database '%s' is already attached", name);
        return false;
    }

    if (!__attach(db, name, dbpath)) {
        LOG_ERROR("DB", "Failed to attach database named '%s' with "
                  "path '%s'", name, dbpath);
        return false;
    }

    LOG_INFO("DB", "Database '%s' attached", name);
    return true;
}

bool
database_detach(volatile sqlite3 *db, const char *name)
{
    RETURN_UNLESS(strcasecmp(PROTECTED_DATABASE, name), " The database "
                  PROTECTED_DATABASE " is protected, can't detach it", false);

    if (is_sql_str_invalid(name)) {
        LOG_ERROR("DB", "The database name '%s' is invalid", name);
        return false;
    }

    if (!__is_attached(db, name)) {
        LOG_ERROR("DB", "The database '%s' is not attached", name);
        return false;
    }

    if (!__detach(db, name)) {
        LOG_ERROR("DB", "Failed to detach database named %s", name);
        return false;
    }

    LOG_INFO("DB", "Database '%s' detached", name);
    return true;
}

bool
database_init(const char *dbpath)
{
    sqlite3 *db;
    if (SQLITE_OK != sqlite3_open(dbpath, &db))
        goto error;
    SQLITE_EXEC(db, (const char *) ___src_database_disk_sql, error);
    LOG_INFO("DB", "Initialized the 'disk' database successfully, "
             "path was '%s'", dbpath);
    return true;
error:
    LOG_ERROR("DB", "Failed to init the 'disk' database, path was '%s'",
              dbpath);
    return false;
}
